use crate::player::PlayerMarker;
use bevy::{
  diagnostic::{Diagnostics, FrameTimeDiagnosticsPlugin},
  prelude::*,
};
use bevy_rapier3d::{physics::RigidBodyHandleComponent, rapier::dynamics::RigidBodySet};

pub struct HudPlugin;

impl Plugin for HudPlugin {
  fn build(&self, app: &mut AppBuilder) {
    app
      .add_plugin(FrameTimeDiagnosticsPlugin::default())
      .add_startup_system(setup.system())
      .add_system(fps_update_system.system())
      .add_system(speed_update_system.system());
  }
}

struct FpsText;
struct SpeedText;

fn fps_update_system(diagnostics: Res<Diagnostics>, mut query: Query<&mut Text, With<FpsText>>) {
  for mut text in query.iter_mut() {
    if let Some(fps) = diagnostics.get(FrameTimeDiagnosticsPlugin::FPS) {
      if let Some(average) = fps.value() {
        text.sections.first_mut().unwrap().value = format!("{:.0} fps", average);
      }
    }
  }
}

fn speed_update_system(
  rigid_body_set: Res<RigidBodySet>,
  mut text_query: Query<&mut Text, With<SpeedText>>,
  player_query: Query<&RigidBodyHandleComponent, With<PlayerMarker>>,
  time: Res<Time>,
) {
  for player in player_query.iter() {
    // TODO: would be better to calculate this in a function in player somehow
    let player_body = rigid_body_set.get(player.handle());
    let player_velocity = player_body.unwrap().linvel();
    let player_velocity_2d = Vec3::new(player_velocity.x, 0.0, player_velocity.z);
    let player_speed_2d = player_velocity_2d.length() / time.delta_seconds();

    for mut text in text_query.iter_mut() {
      text.sections.first_mut().unwrap().value = format!("{:.0} m/s", player_speed_2d);
    }
  }
}

fn setup(mut commands: Commands, asset_server: Res<AssetServer>) {
  commands
    // UI camera
    .spawn_bundle(UiCameraBundle::default());
    // fps display
    commands.spawn_bundle(TextBundle {
      style: Style {
        align_self: AlignSelf::FlexEnd,
        position_type: PositionType::Absolute,
        position: Rect {
          top: Val::Px(5.0),
          right: Val::Px(5.0),
          ..Default::default()
        },
        ..Default::default()
      },
      text: Text::with_section(
        "".to_string(),
        TextStyle {
          font: asset_server.load("fonts/FiraMono-Medium.ttf"),
          font_size: 16.0,
          color: Color::WHITE,
          ..Default::default()
        },
        TextAlignment::default()
      ),
      ..Default::default()
    })
    .insert(FpsText);
    // speed display
    commands.spawn_bundle(TextBundle {
      style: Style {
        // TODO: dont know how to properly center this horizontally
        align_self: AlignSelf::Center,
        position_type: PositionType::Relative,
        position: Rect {
          top: Val::Px(50.0),
          right: Val::Px(-480.0),
          ..Default::default()
        },
        ..Default::default()
      },
      text: Text::with_section(
        "speed".to_string(),
        TextStyle {
          font: asset_server.load("fonts/FiraMono-Medium.ttf"),
          font_size: 16.0,
          color: Color::rgba(1.0, 1.0, 1.0, 0.5),
          ..Default::default()
        },
        TextAlignment { vertical: VerticalAlign::Center, horizontal: HorizontalAlign::Center},
    ), ..Default::default()
    })
    .insert(SpeedText);
}
