use bevy::{input::mouse::MouseMotion, prelude::*};
use bevy_rapier3d::{
  na::{clamp, wrap, Matrix3x1},
  physics::{ColliderHandleComponent, RigidBodyHandleComponent},
  rapier::{
    dynamics::{RigidBody, RigidBodyBuilder, RigidBodySet},
    geometry::{ColliderBuilder, NarrowPhase},
  },
};
use std::f32::consts::{FRAC_PI_2, PI};

pub struct PlayerPlugin;

impl Plugin for PlayerPlugin {
  fn build(&self, app: &mut AppBuilder) {
    app
      .add_system(player_movement_system.system())
      .add_system(player_look_system.system())
      .add_startup_system(init_player.system());
  }
}

#[derive(Default)]
pub struct PlayerMarker;

#[derive(Default, Debug)]
struct CameraData {
  yaw: f32,
  pitch: f32,
}

#[derive(Bundle)]
pub struct PlayerBundle {
  pub transform: Transform,
  pub global_transform: GlobalTransform,
  player_marker: PlayerMarker,
  rigid_body: RigidBodyBuilder,
  collider: ColliderBuilder,
}

impl Default for PlayerBundle {
  fn default() -> Self {
    PlayerBundle {
      rigid_body: RigidBodyBuilder::new_dynamic()
        .translation(0.0, 3.0, 0.0)
        .lock_rotations(),
      collider: ColliderBuilder::cylinder(PLAYER_HEIGHT / 2., PLAYER_WIDTH).friction(0.0),
      transform: Default::default(),
      global_transform: Default::default(),
      player_marker: Default::default(),
    }
  }
}

const PLAYER_SPEED: f32 = 320.0;
const MOUSE_SENSITIVITY: f32 = 0.5;
const PLAYER_HEIGHT: f32 = 1.8;
const PLAYER_WIDTH: f32 = 0.25;

fn init_player(mut commands: Commands) {
  commands
    .spawn_bundle(PlayerBundle::default())
    .with_children(|parent| {
      parent
        .spawn_bundle(PerspectiveCameraBundle {
          transform: Transform::from_translation(Vec3::new(0.0, PLAYER_HEIGHT / 2., 0.0)),
          ..Default::default()
        })
        .insert(CameraData::default());
    });
}

// TODO: this is incomplete, currently just checks that the player is touching anything
fn is_on_ground(
  narrow_phase: &Res<NarrowPhase>,
  collider_handle: &ColliderHandleComponent,
) -> bool {
  if let Some(n1) = narrow_phase.contacts_with(collider_handle.handle()) {
    for _n2 in n1 {
      return true;
    }
  }
  return false;
}

fn player_movement_system(
  time: Res<Time>,
  mut rigid_body_set: ResMut<RigidBodySet>,
  keyboard_input: Res<Input<KeyCode>>,
  narrow_phase: Res<NarrowPhase>,
  camera_transform_query: Query<&Transform, With<CameraData>>,
  mut player_physics_query: Query<
    (&RigidBodyHandleComponent, &ColliderHandleComponent),
    With<PlayerMarker>,
  >,
) {
  for camera_transform in camera_transform_query.iter() {
    for (body_handle, collider_handle) in player_physics_query.iter_mut() {
      let is_on_ground = is_on_ground(&narrow_phase, collider_handle);
      if !is_on_ground {
        continue;
      }
      let body: &mut RigidBody = rigid_body_set.get_mut(body_handle.handle()).unwrap();
      if keyboard_input.pressed(KeyCode::Space) {
        body.apply_impulse(Matrix3x1::new(0.0, 0.5, 0.0), true);
      }

      let axis_strafe = movement_axis(&keyboard_input, KeyCode::D, KeyCode::A);
      let axis_straight = movement_axis(&keyboard_input, KeyCode::S, KeyCode::W);

      let rotation = camera_transform.rotation;
      let mut speed_2d: Vec3 =
        (strafe_vector(&rotation) * axis_strafe) + (straight_vector(&rotation) * axis_straight);
      if speed_2d.length() != 0.0 {
        speed_2d = speed_2d.normalize() * PLAYER_SPEED;
      }

      let velocity_2d = speed_2d * time.delta_seconds();

      let velocity_3d = Matrix3x1::new(velocity_2d.x, body.linvel().y, velocity_2d.z);
      body.set_linvel(velocity_3d, true);
    }
  }
}

fn straight_vector(rotation: &Quat) -> Vec3 {
  let v = rotation.mul_vec3(Vec3::Z);
  Vec3::new(v.x, 0.0, v.z).normalize()
}

fn strafe_vector(rotation: &Quat) -> Vec3 {
  let forward = straight_vector(rotation);
  Quat::from_rotation_y(FRAC_PI_2)
    .mul_vec3(forward)
    .normalize()
}

fn movement_axis(input: &Res<Input<KeyCode>>, plus: KeyCode, minus: KeyCode) -> f32 {
  let mut axis = 0.0;
  if input.pressed(plus) {
    axis += 1.0;
  }
  if input.pressed(minus) {
    axis -= 1.0;
  }
  axis
}

fn player_look_system(
  time: Res<Time>,
  mut mouse_event_reader: EventReader<MouseMotion>,
  mut query: Query<(&mut Transform, &mut CameraData)>,
) {
  let mut delta: Vec2 = Vec2::ZERO;
  for event in mouse_event_reader.iter() {
    delta += event.delta;
  }
  if delta.is_nan() {
    return;
  }

  for (mut transform, mut data) in query.iter_mut() {
    data.yaw -= delta.y * MOUSE_SENSITIVITY * time.delta_seconds();
    data.pitch -= delta.x * MOUSE_SENSITIVITY * time.delta_seconds();
    // NOTE: if we just use FRAC_PI_2 here, for some reason it results in a blank screen when
    //       looking up/down
    let half_pi = FRAC_PI_2 - 0.01;
    data.yaw = clamp(data.yaw, -half_pi, half_pi);
    data.pitch = wrap(data.pitch, -PI, PI);

    let quat_yaw = Quat::from_axis_angle(Vec3::X, data.yaw);
    let quat_pitch = Quat::from_axis_angle(Vec3::Y, data.pitch);

    let mut orientation = Quat::IDENTITY;
    orientation = orientation * quat_yaw;
    orientation = quat_pitch * orientation;

    transform.rotation = orientation.normalize();
  }
}
